from typing import Tuple
import numpy as np

from ..misc.utils import isType


def firstOrderSpline(x: np.ndarray, y: np.ndarray):
    """
    Computes the coefficients of a first order spline, according to the formulas:

    .. math::
        a_k=\\frac{y_{k+1}-y_k}{x_{k+1}-x_k}
    .. math::
        b_k=y_k-a_k*x_k

    :param x: function arguments (x axis)
    :param y: function values (y axis)

    :return: the coefficients of the linear function in the tuple of shape (a0, a1)"""

    if not isType(x, np.ndarray) or not isType(y, np.ndarray) or len(x) != len(y):
        return None

    a = []
    b = []

    for k in range(len(x) - 1):
        a.append((y[k + 1] - y[k]) / (x[k + 1] - x[k]))
        b.append(y[k] - a[k] * x[k])

    return (a, b)


def cubicSpline(x: np.ndarray, y: np.ndarray) -> Tuple[float, float, float, float]:
    """
    Computes the coefficients of a third order (cubic) spline

    :param x: function arguments (x axis)
    :param y: function values (y axis)

    :return: the coefficients in the tuple of shape (a0, a1, a2, a3)"""

    if not isType(x, np.ndarray) or not isType(y, np.ndarray) or len(x) != len(y):
        return np.nan, np.nan, np.nan, np.nan

    h = np.zeros(len(y))
    d = np.zeros(len(y))
    lmb = np.zeros(len(y))
    rho = np.zeros(len(y))
    m = np.zeros(len(y))
    b3 = np.zeros(len(y) - 1)
    b2 = np.zeros(len(y) - 1)
    b1 = np.zeros(len(y) - 1)
    b0 = np.zeros(len(y) - 1)

    tgtLen = len(x) - 1

    for k in range(tgtLen):
        h[k] = x[k + 1] - x[k]
        d[k] = (y[k + 1] - y[k]) / h[k]

    for k in range(tgtLen):
        lmb[k + 1] = h[k + 1] / (h[k] + h[k + 1])
        rho[k + 1] = h[k] / (h[k] + h[k + 1])

    for k in range(tgtLen - 1):
        m[k + 1] = 3 * (d[k + 1] - d[k]) / (h[k] + h[k + 1]) - \
            m[k] * rho[k + 1] / 2 - m[k + 2] * lmb[k + 1] / 2

    for k in range(tgtLen):
        b0[k] = y[k]
        b1[k] = d[k] - h[k] * (2 * m[k] + m[k + 1]) / 6
        b2[k] = m[k] / 2
        b3[k] = (m[k + 1] - m[k]) / 6 * h[k]

    xCut = x[:-1]
    a3 = b3
    a2 = b2 - 3 * b3 * xCut
    a1 = b1 - 2 * b2 * xCut + 3 * b3 * xCut**2
    a0 = b0 - b1 * xCut + b2 * xCut**2 - b3 * xCut**3

    return (a0, a1, a2, a3)



from math import pi

import numpy as np


def cylinderArea(r: float, h: float):
    """
    Computes the total area of a cylinder

    :param r: cylinder base radius
    :param h: cylinder height

    :return: cylinder area
    :rtype: float
    """

    if r > 0 and h > 0:
        return pi * pow(r, 2) * 2 + 2 * pi * r * h
    else:
        return np.nan


def fibonacci(n: int):
    """
    Computes the first n Fibonacci's sequence elements

    Parameters:
    :param n: the amount of elements

    :return: n first Fibonacci's sequence elements
    :rtype: np.ndarray
    """

    if n <= 0 or isinstance(n, float):
        return None
    else:
        arr = np.ndarray(shape=(1, n), dtype=int)

        for i in range(0, n):
            if i == 0 or i == 1:
                arr[0][i] = 1
            else:
                arr[0][i] = arr[0][i - 1] + arr[0][i - 2]

        return arr



from typing import Optional
import numpy as np

from ..misc.utils import isType


def chebyshevNodes(n: int = 10) -> Optional[np.ndarray]:
    """Creates a vector of Chebyshev nodes in the shape of (n+1, )

    :param n: amount of Chebyshev nodes, a positive integer

    :returns: the vector of chebyshev nodes
    :rtype: np.ndarray
    """

    if not isType(n, int) or n < 0:
        return None

    # return np.cos(np.linspace(1, n, n) * np.pi / n)
    return np.array([np.cos(i * np.pi / n) for i in range(0, n + 1)])



from typing import Any, Union, List, cast

import numpy as np


def isType(val: Any, typeArr: Union[List[Any], Any]) -> bool:
    """Helper that checks if the supplied value `val` is of a specific type or of a type present in `typeArr`"""

    if isinstance(typeArr, List):
        anyMatched = False

        for type in typeArr:
            if isinstance(val, type):
                anyMatched = True

        return anyMatched
    else:
        if not isinstance(val, typeArr):
            return False
    return True


def absoluteError(v: Union[int, float, List, np.ndarray], v_approx: Union[int, float, List, np.ndarray]) -> Union[int, float, np.ndarray]:
    """
    Calculates an absolute error of two arguments, either scalars, or vectors

    :param v: precise value
    :param v_approx: approximate value

    :return: absolute error value or NaN if input data is ill-formed
    """

    try:
        if((not isType(v, [int, float, List, np.ndarray])) or (not isType(v_approx, [int, float, List, np.ndarray]))):
            return np.nan

        if isinstance(v, List):
            v = np.array(v)

        if isinstance(v_approx, List):
            v_approx = np.array(v_approx)

        if isinstance(v, np.ndarray):
            return np.abs(cast(np.ndarray, v) - v_approx)
        elif isinstance(v_approx, np.ndarray):
            return np.abs(cast(np.ndarray, v) - v_approx)
        else:
            return np.abs(cast(Union[int, float], v) - cast(Union[int, float], v_approx))
    except:
        return np.nan


def relativeError(v: Union[int, float, List, np.ndarray], v_approx: Union[int, float, List, np.ndarray]) -> Union[int, float, np.ndarray]:
    """
    Calculates a relative error of two arguments, either scalars, or vectors

    :param v: precise value
    :param v_approx: approximate value

    :return: relative error value or NaN if input data is ill-formed
    """

    try:
        if not isType(v, [int, float, List, np.ndarray]):
            return np.nan

        if isType(v, [int, float]):
            if v == 0:
                return np.nan
        else:
            hasZero = False

            for val in cast(Union[List, np.ndarray], v):
                if val == 0:
                    hasZero = True

            if hasZero:
                return np.nan

        res = absoluteError(v, v_approx)

        if res is np.nan:
            return np.nan
        else:
            return res / np.abs(v)
    except:
        return np.nan



from typing import Callable, Tuple, Union

import numpy as np

from .utils import isType


def exponential(x: Union[int, float], n: int) -> float:
    """
    Computes an approximation of e^x

    :param x: x
    :param k: k

    :return: the approximated value
    """

    acc = 0

    try:
        if isType(x, [int, float]) and isType(n, int) and n >= 0:
            for i in range(0, n):
                acc += pow(x, i) / np.math.factorial(i)

            return acc
        else:
            return np.nan
    except OverflowError:
        return acc


def coskx1(k: int, x: Union[int, float]) -> float:
    """
    Computes an approximation of cos(kx) using the formula:

    .. math::
        cos((m+1)x) = 2cos(x) * cos(mx) - cos((m-1)x)

    :param x: x
    :param k: k

    :return: the approximated value
    """

    try:
        if isType(k, int) and k >= 0 and isType(x, [int, float]):
            m1 = k - 1
            m2 = m1 - 1

            if m1 >= 1:
                if m1 == 1:
                    e1 = np.math.cos(x)
                else:
                    e1 = coskx1(m1, x)
            else:
                e1 = 1

            if m2 >= 1:
                if m2 == 1:
                    e2 = np.math.cos(x)
                else:
                    e2 = coskx1(m2, x)
            else:
                e2 = 1

            if k == 0:
                return np.math.cos(0)
            elif k == 1:
                return np.math.cos(x)
            else:
                return 2 * np.math.cos(x) * e1 - e2
        else:
            return np.nan
    except:
        return np.nan


def cossinkx2(k: int, x: Union[int, float]) -> Tuple[float, float]:
    """
    Computes an approximation of cos(kx) & sin(kx) using the formulas:

    .. math::
        cos(mx) = cosx \\cdot cos(m-1)x - sinx \\cdot sin(m-1)x
    .. math::
        sin(mx) = sinx \\cdot cos(m-1)x + cosx \\cdot sin(m-1)x

    :param x: x
    :param k: k

    :return: approximated cos(kx) and sin(kx) values
    """

    try:
        if isType(k, int) and k >= 0 and isType(x, [int, float]):
            if k == 0:
                return (np.math.cos(0), np.math.sin(0))
            elif k == 1:
                return (np.math.cos(x), np.math.sin(x))
            else:
                (cosKLess1, sinKLess1) = cossinkx2((k - 1), x)
                return (np.math.cos(x) * cosKLess1 - np.math.sin(x) * sinKLess1, np.math.sin(x) * cosKLess1 + np.math.cos(x) * sinKLess1)
        else:
            return np.nan, np.nan
    except:
        return np.nan, np.nan


def pi(n: int) -> float:
    """Computes an approximate value of Pi, using the formula:

    .. math::
        \\sum_{n=1}^{\\infty} \\cfrac{1}{n^2} = \\cfrac{1}{6} \\pi^2


    :param n: the amount of sum elements

    :return: the approximated value of pi
    """

    if isType(n, int) and n > 0:
        acc = 0

        for i in range(1, n + 1):
            acc += 1 / pow(i, 2)

        return np.math.sqrt(acc * 6)
    else:
        return np.nan


def approxZeroBisection(a: Union[int, float], b: Union[int, float], f: Callable[[float], float], epsilon: float, iteration: int) -> Tuple[float, int]:
    """
    Approximates the solution to f(x) = 0 in range [a, b] using bisection method

    :param a: left range bound
    :param b: right range bound
    :param f: callable function of variable x
    :param epsilon: the desired precision (stop condition)
    :param iteration: maximum iterations count limit (to prevent an infinite loop)

    :raises ValueError: when input types are wrong or f has same signs on both range bounds

    :return: a tuple containing (in order): an approximate solution x, iterations made count 
    """

    if not isType(a, [int, float]) or not isType(b, [int, float]) or not isType(f, Callable) or not isType(epsilon, float) or not isType(iteration, int) or epsilon < 0 or iteration < 0 or b < a:
        raise ValueError("Wrong input types")

    if np.sign(f(a)) * np.sign(f(b)) >= 0:
        raise ValueError(
            "Function f has to be of different signs at the edges of the range [a, b]")

    i = 0
    c = (a + b) / 2
    while np.abs(f(c)) > epsilon and i < iteration:
        c = (a + b) / 2

        if np.sign(f(a)) != np.sign(f(c)):
            b = c
        else:
            a = c

        i += 1

    return c, i - 1


def approxZeroSecant(a: float, b: float, f: Callable[[float], float], epsilon: float, iteration: int) -> Tuple[float, int]:
    """
    Approximates the solution to f(x) = 0 in range [a, b] using secant method

    :param a: left range bound
    :param b: right range bound
    :param f: callable function of variable x
    :param epsilon: the desired precision (stop condition)
    :param iteration: maximum iterations count limit (to prevent an infinite loop)

    :raises ValueError: when input types are wrong or f has same signs on both range bounds

    :return: a tuple containing (in order): an approximate solution x, iterations made count 
    """

    if not isType(a, [int, float]) or not isType(b, [int, float]) or not isType(f, Callable) or not isType(epsilon, float) or not isType(iteration, int) or epsilon < 0 or iteration < 0 or b < a or np.sign(f(a)) * np.sign(f(b)) >= 0:
        raise ValueError("Wrong input types")

    if np.sign(f(a)) * np.sign(f(b)) >= 0:
        raise ValueError(
            "Function f has to be of different signs at the edges of the range [a, b]")

    f_a = f(a)
    f_b = f(b)
    x0 = 0
    i = 0
    while iteration > -1 and (abs(a - b) > epsilon):
        i += 1

        if abs(f_a - f_b) < epsilon:
            break

        x0 = a - f(a) * (a - b) / (f(a) - f(b))
        f0 = f(x0)

        if abs(f0) < epsilon:
            break

        if f(a) * f0 < 0:
            b = x0

        elif f(b) * f0:
            a = x0

        elif f0 == 0:
            return x0, iteration

        iteration -= 1

    return x0, i - 1


def approxZeroNewton(f: Callable[[float], float], df: Callable[[float], float], ddf: Callable[[float], float], a: Union[int, float], b: Union[int, float], epsilon: float, iteration: int) -> Tuple[float, int]:
    """
    Approximates the solution to f(x) = 0 in range [a, b] using Newton's method

    :param f: callable function of variable x
    :param df: callable derivate of f (f')
    :param ddf: callable second-order derivate of f (f'')
    :param a: left range bound
    :param b: right range bound
    :param epsilon: the desired precision (stop condition)
    :param iteration: maximum iterations count limit (to prevent an infinite loop)

    :raises ValueError: when input types are wrong or either df or dff has same different on both range bounds

    :return: a tuple containing (in order): an approximate solution x, iterations made count 
    """

    if not isType(a, [int, float]) or not isType(b, [int, float]) or not isType(f, Callable) or not isType(df, typing.Callable) \
            or not isType(ddf, Callable) or not isType(epsilon, float) or not isType(iteration, int) or epsilon < 0 or iteration < 0 or b < a or np.sign(f(a)) * np.sign(f(b)) >= 0:
        raise ValueError("Wrong input types")

    if np.sign(df(a)) * np.sign(df(b)) <= 0:
        raise ValueError(
            "Df has to be of different signs at the edges of range [a, b]")

    if np.sign(ddf(a)) * np.sign(ddf(b)) <= 0:
        raise ValueError(
            "Ddf has to be of different signs at the edges of range [a, b]")

    x0 = a
    for x0 in [a, b]:
        if(f(x0) * ddf(x0) > 0):
            # initial condition found
            break

    i = 0
    while np.abs(f(x0)) > epsilon and i < iteration:
        x0 = x0 - f(x0) / df(x0)

        i += 1

    return x0, i



import numpy as np

from ..misc.utils import isType


def isMatrixDiagDominant(A: np.ndarray) -> bool:
    """
    Checks if matrix is strictly diagonally dominated

    :param A: the matrix to be checked

    :raises ValueError: If the supplied np.ndarray does not have exactly 2 dimensions or it is not square

    :return: whether the matrix is strictly diagonally dominated
    """

    if not isType(A, np.ndarray) or len(A.shape) != 2:
        raise ValueError("A matrix must have exactly 2 dimensions")

    m, shouldBeM = A.shape
    if m != shouldBeM:
        raise ValueError(
            f"The matrix has to be square, while a shape of ({m}, {shouldBeM}) was supplied")

    diagVals = np.diag(np.abs(A))
    rowSumsExclDiag = np.sum(np.abs(A), axis=1) - diagVals

    if np.all(diagVals > rowSumsExclDiag):
        return True
    else:
        return False


def isMatrixSymmetric(A: np.ndarray) -> bool:
    """
    Checks if a square matrix is symmetric

    :param A: the matrix to be checked

    :raises ValueError: If the supplied np.ndarray does not have exactly 2 dimensions or it is not square

    :return: whether the matrix is symmetric
    """

    if not isType(A, np.ndarray) or len(A.shape) != 2:
        raise ValueError("A matrix must have exactly 2 dimensions")

    m, shouldBeM = A.shape
    if m != shouldBeM:
        raise ValueError(
            f"The matrix has to be square, while a shape of ({m}, {shouldBeM}) was supplied")

    for i in range(0, m):
        for j in range(0, round(len(A[i]) / 2)):
            if i != j and A[i][j] != A[j][i]:
                return False

    return True



from typing import List, Tuple, cast
import numpy as np
import scipy.linalg as spLinalg

from ..misc.utils import isType
from .classification import isMatrixDiagDominant


def svdAndReconstruction(A: np.ndarray, singularValues: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Generates an SVD decomposition of matrix A & reconstructs
    the matrix using these components and supplied singular values of matrix A

    :param A: matrix A of shape (m, m)
    :param singularValues: singular values vector of shape (m, 1)

    :raises ValueError: When inputs' types are incorrect or either shape is invalid

    :returns: (U, S, V, M), where U, S & V are SVD decomposition components & M is the reconstruction of A
    """

    if not isType(A, np.ndarray) or not isType(singularValues, np.ndarray):
        raise ValueError("Wrong input types")

    m, shouldBeM = A.shape
    if m != shouldBeM or m != singularValues.shape[0]:
        raise ValueError("Wrong input shapes")

    U, S, V = np.linalg.svd(A)

    return U, S, V, (U * singularValues) @ V


def eqSysRectToSquare(A: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Transforms an equation system with a rectangular matrix of coefficients, to an equation system with a square matrix
    Note: both the matrix & the vector returned will be different from the inputs

    :param A: the rectangular coefficients matrix of shape (m, n)
    :param b: vector b of shape (m, 1), containg the right hand side coefficients

    :raises ValueError: When inputs' types are incorrect or either shape is invalid

    :return: tuple containing a square matrix of shape size (n, n) & a modified vector b of shape (n, 1)
    """

    if not isType(A, np.ndarray) or not isType(b, np.ndarray):
        raise ValueError("Wrong input types")

    m, n = A.shape
    if m != b.shape[0]:
        raise ValueError("Wrong input shapes")

    At = np.transpose(A)

    semiNewA = At @ A
    newB = At @ b

    return semiNewA, newB


def eqSysLeastSquaresAndSolve(A: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Transforms an equation system using the least squares method according to the formula:

    .. math::
        A^T Ax = A^T b

    and solves the equation system using sp.linalg.solve

    :param A: a square coefficients matrix A of shape (m, m)
    :param b: vector of right hand side coefficients b of shape (m, 1)

    :return: a tuple containing (in order): vector of eq sys solutions x, transformed matrix A^T A, transformed vector A^T b
    """

    AtA, Atb = eqSysRectToSquare(A, b)

    x = np.linalg.solve(AtA, Atb)

    return x, AtA, Atb


def eqSysQRAndSolve(A: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Transforms an equation system using QR decomposition and solves the equation system using sp.linalg.solve_triangular

    :param A: a square coefficients matrix A of shape (m, m)
    :param b: vector of right hand side coefficients b of shape (m, 1)

    :return: a tuple containing (in order): vector of eq sys solutions x, Q, R where Q & R are QR decomposition components
    """

    Q, R = cast(Tuple[np.ndarray, np.ndarray], np.linalg.qr(A, mode='reduced'))

    Qb = Q.T @ b
    x = spLinalg.solve_triangular(R, Qb)

    return x, Q, R


def eqSysSVDAndSolve(A: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Transforms an equation system using SVD decomposition and solves the equation system using sp.linalg.solve

    :param A: a square coefficients matrix A of shape (m, m)
    :param b: vector of right hand side coefficients b of shape (m, 1)

    :return: a tuple containing (in order): vector of eq sys solutions x, U, S, V where U, S & V are SVD decomposition components
    """

    U, S, V = np.linalg.svd(A)

    c = U.T @ b
    w = np.linalg.solve(np.diag(S), c)
    x = V.T @ w

    return x, U, S, V


def eqSysLUAndSolveJacobi(A: np.ndarray, b: np.ndarray, x_init: np.ndarray,
                          epsilon: float = 1e-8, maxiter: int = 100, checkNecessaryCondition: bool = False) -> Tuple[np.ndarray, int]:
    """
    Transforms an equation system using LU decomposition and solves the equation system using Jacobi's method

    :param A: a square coefficients matrix A of shape (m, m)
    :param b: vector of right hand side coefficients b of shape (m, 1)
    :param x_init: initial solution of shape (m, 1)
    :param epsilon: the desired solution precision
    :param maxiter: a maximum iterations limit (to prevent an infinite loop), a positive integer

    :raises ValueError: If either input is of invalid type, matrix A does not have exactly 2 dimensions, vector b is of invalid shape,\
        b's first dimension does not match A's, matrix A is not square, x_init has an invalid shape, epsilon or maxiter are not positive integers,\
            or matrix A is not strictly diagonally dominant (in such a case Jacobi's method would not converge)

    :return: a tuple containing (in order): eq sys solution x of shape (m, 1), L, U, iter, where L & U are LU decomposition components \
        and iter is the number of iterations passed
    """

    if not isType(A, np.ndarray) or len(A.shape) != 2 or not isType(b, np.ndarray) or len(b.shape) > 2 \
            or not isType(x_init, np.ndarray) or not isType(epsilon, float) or not isType(maxiter, int) or maxiter < 0 or epsilon < 0:
        raise ValueError("Invalid input values or types")

    m, shouldBeM = A.shape
    if m != shouldBeM:
        raise ValueError("Matrix A is not square")

    if len(b.shape) == 1:
        if b.shape[0] != m:
            raise ValueError("Vector b's first dimension is of invalid length")
    else:
        shouldAlsoBeM, shouldBe1 = b.shape
        if shouldBe1 != 1 or m != shouldAlsoBeM:
            raise ValueError(
                "Vector b is of invalid shape")

    xInitShouldBeM, xInitShouldBe1 = x_init.shape
    if xInitShouldBe1 != 1 or xInitShouldBeM != m:
        raise ValueError("Vector xinit is of invalid shape")

    # necessary convergence condition of Jacobi's method
    if checkNecessaryCondition and not isMatrixDiagDominant(A):
        raise ValueError(
            "Matrix A is not strictly diagonally dominant, therefore Jacobi's method is not convergent for it")

    diagMat = np.diag(np.diag(A))
    LU = A - diagMat

    x = x_init
    invDiagMat = np.diag(1 / np.diag(diagMat))
    residuals = []
    for i in range(maxiter):
        x_new = invDiagMat @ (b - LU @ x)
        rnorm = np.linalg.norm(x_new - x)

        residuals.append(rnorm)

        if rnorm < epsilon:
            return x_new, i + 1

        x = x_new

    return x, maxiter


def frobeniusFromPolyCoeffs(coeffs: List[float]) -> np.ndarray:
    """
    Generates a Frobenius matrix for a given list of polynomial coefficients, ordered from the one next to the highest power of x, to the one next to the lowest (that is, just a scalar)

    For example: w(x) = 5x^3 + 4x - 3 => coeffs = [5, 0, 4, -3]

    .. math::
        \\setcounter{MaxMatrixCols}{20}
        \\begin{vmatrix}
            1     &            &           &                  \\\\
                  &  \\ddots   &           &                  \\\\
                  &            &     1     &                  \\\\
            -a_0  &    -a_1    &  \\dots   &    -a_{n-1}
        \\end{vmatrix}

    :raises ValueError: when the input is not a list, or the first coefficient is equal to 0 (and cannot be a divisor then)

    :returns: the Frobenius array corresponding to the polynomial with given coefficients
    """

    if not isType(coeffs, List):
        raise ValueError("Wrong input type")

    if coeffs[0] == 0:
        raise ValueError(
            "The highest power's coefficient is equal to 0 (therefore cannot be a divisor)")

    aBiggestPower, *coeffs = coeffs

    coeffs = np.array(coeffs) / aBiggestPower

    coeffsLen = len(coeffs)

    frob = np.zeros(shape=(coeffsLen, coeffsLen))

    # fill ones
    for i in range(coeffsLen - 1):
        frob[i][i + 1] = 1

    # fill last row
    lastRowInd = coeffsLen - 1
    for i in range(coeffsLen):
        frob[lastRowInd][len(frob[lastRowInd]) - i - 1] = -coeffs[i]

    return frob



from typing import Any, List, Union, cast
import numpy as np

from ..misc.utils import isType


def maxNorm(xr: Union[int, float, List, np.ndarray], x: Union[int, float, List, np.ndarray]) -> float:
    """
    Computes the max norm (sometimes called the 'infinity norm') of scalars or vectors; both arguments' dimensions must be compatible

    :param xr: precise value, either a scalar or a vector of shape (n, 1)
    :param x: approximate value, either a scalar or a vector of shape (n, 1)

    :return: the value of the norm, max(abs(xr - x))
    """

    if not isType(xr, [int, float, List, np.ndarray]) or not isType(x, [int, float, List, np.ndarray]):
        return np.nan

    if (isType(xr, [int, float]) and not isType(x, [int, float])) or (isType(x, [int, float]) and not isType(xr, [int, float])):
        return np.nan

    xNp: np.ndarray
    if isType(x, [List]):
        xNp = np.array(x)
    else:
        xNp = cast(Any, x)

    xrNp: np.ndarray
    if isType(xr, [List]):
        xrNp = np.array(xr)
    else:
        xrNp = cast(Any, xr)

    if isType(xrNp, [int, float]) or isType(xNp, [int, float]):
        return np.max(np.abs(xrNp - xNp))
    else:
        if xrNp.shape == xNp.shape:
            return np.max(np.abs(xrNp - xNp))
        else:
            print(xrNp.shape, xNp.shape)
            return np.nan


def residualNorm(A: np.ndarray, x: np.ndarray, b: np.ndarray) -> float:
    """
    Computes the matrix norm of the residues of an equation in a form of Ax = b


    :param A: matrix A (m, m) containing the equation's coefficients
    :param x: vector x (m, 1) containing the equation's solutions
    :param b: vector b (m, 1) containing the equation's right hand side's coefficients

    :return: matrix norm value of the equation's residues"""

    if not isType(A, np.ndarray) or not isType(x, np.ndarray) or not isType(b, np.ndarray):
        return np.nan

    m, n = A.shape
    if m != n or m != b.shape[0] or m != x.shape[0]:
        return np.nan

    return np.linalg.norm(b - A @ x)
